#include <nucleus.h>
#include <libndls.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

typedef int CREATETASK(void *task_ptr, const char *name,
	void (*task_entry)(int, char**), int argc, const char **argv,
	void *stack_address, int stack_size,
	int priority, int time_slice, int preempt, int auto_start);

void uart_putchar(char c)
{
        volatile unsigned *line_status_reg = (unsigned*)0x90020018;
        volatile unsigned *xmit_holding_reg = (unsigned*)0x90020000;
        while(!(*line_status_reg & 0b10000000));
        *xmit_holding_reg = c;
}

void uart_puts(const char *c)
{
	while(*c)
		uart_putchar(*c++);
}

void uart_printf(const char *fmt, ...)
{
	char s[1024];
	va_list va;
	va_start(va, fmt);
	vsprintf(s, fmt, va);  
	va_end(va);
}

uint16_t *screen, *buffer;
class Flake {
	public:
		void spawn()
		{
			x = rand() % 320;
			y = (rand() % 100) - 100;
			size = rand() % 3;
		}
		void logic()
		{
			y += rand() % 3;
			x += (rand() % 2) - 1;
			if(y >= 240)
				spawn();
		}
		void render()
		{
			for(int x = this->x - size; x < this->x + size; ++x)
				for(int y = this->y - size; y < this->y + size; ++y)
					if(x >= 0 && x < 320 && y >= 0 && y < 240)
						buffer[x + 320*y] = 0xFFFF;
		}

	private:
		int16_t x, y, size;
};

Flake flakes[100];

extern "C" void snow(int, char**)
{
	void *old_screen = SCREEN_BASE_ADDRESS;
	SCREEN_BASE_ADDRESS = screen;
	while(true)
	{
		memcpy(buffer, old_screen, 320*240*2);
		for(auto &flake : flakes)
		{
			flake.logic();
			flake.render();
		}
		memcpy(screen, buffer, 320*240*2);
	}
}

int main()
{
	unsigned int *i = new unsigned int;
	srand(*i);
	delete i;
	screen = new uint16_t[320*240];
	buffer = new uint16_t[320*240];
	void *task = malloc(0x1000);
	void *stack = reinterpret_cast<char*>(malloc(0x4000));
	printf("Task at %p (stack %p)\n", task, stack);
	static const unsigned createtask_addrs[] = 
	{
                0x102F0A2C,//non-CAS 3.1
		0x102F1158,//CAS 3.1
		0x102F0498,//non-CAS CX 3.1
		0x102F0C28,//CAS CX 3.1
		0x102F0C28,//CM-C 3.1 (copied from CAS CX 3.1, not sure???)
		0x102DC134,//CAS CM-C 3.1
		0x10375634,//non-CAS 3.6
		0x10376074,//CAS 3.6
		0x103750AC,//non-CAS CX 3.6
		0x10375B1C//CAS CX 3.6
	};
#define CreateTask SYSCALL_CUSTOM(createtask_addrs, int, void *task_ptr, const char *name, void (*task_entry)(int, char**), int argc, const char **argv, void *stack_address, int stack_size, int priority, int time_slice, int preempt, int auto_start)
	CreateTask(task, "SNOW", snow, 0, nullptr, stack, 0x4000, 100, 1, 10, 12);

	nl_set_resident();
	return 0;
}

